@@ -8,19 +8,23 @@
     :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
     :license: BSD, see LICENSE for more details.
 """
-from __future__ import with_statement
-
 import os
 import sys
 import thread
 import threading
 import traceback
-from contextlib import contextmanager
 from itertools import chain
 from weakref import ref as weakref
 from datetime import datetime
 
-from logbook.helpers import to_safe_json, parse_iso8601, cached_property, F
+from logbook.helpers import to_safe_json, parse_iso8601, cached_property, \
+     F, _py3
+try:
+    from logbook._speedups import group_reflected_property, \
+         ContextStackManager, StackedObject
+except ImportError:
+    from logbook._fallback import group_reflected_property, \
+         ContextStackManager, StackedObject
 
 
 # make sure to sync these up with _speedups.pyx
@@ -45,16 +49,31 @@ _reverse_level_names = dict((v, k) for (k, v) in _level_names.iteritems())
 _missing = object()
 
 
+# on python 3 we can savely assume that frame filenames will be in
+# unicode, on Python 2 we have to apply a trick.
+if _py3:
+    def _convert_frame_filename(fn):
+        return fn
+else:
+    def _convert_frame_filename(fn):
+        if isinstance(fn, unicode):
+            fn = fn.decode(sys.getfilesystemencoding() or 'utf-8',
+                           'replace')
+        return fn
+
+
 def level_name_property():
     """Returns a property that reflects the level as name from
     the internal level attribute.
     """
+
     def _get_level_name(self):
         return get_level_name(self.level)
+
     def _set_level_name(self, level):
         self.level = lookup_level(level)
-    return property(_get_level_name, _set_level_name, doc=
-        'The level as unicode string')
+    return property(_get_level_name, _set_level_name,
+                    doc='The level as unicode string')
 
 
 def lookup_level(level):
@@ -67,14 +86,6 @@ def lookup_level(level):
         raise LookupError('unknown level name %s' % level)
 
 
-try:
-    from logbook._speedups import group_reflected_property, \
-         ContextStackManager
-except ImportError:
-    from logbook._fallback import group_reflected_property, \
-         ContextStackManager
-
-
 def get_level_name(level):
     """Return the textual representation of logging level 'level'."""
     try:
@@ -86,8 +97,18 @@ def get_level_name(level):
 class ExtraDict(dict):
     """A dictionary which returns ``u''`` on missing keys."""
 
-    def __missing__(self, key):
-        return u''
+    if sys.version_info[:2] < (2, 5):
+        def __getitem__(self, key):
+            try:
+                return dict.__getitem__(self, key)
+            except KeyError:
+                return u''
+    else:
+        def __missing__(self, key):
+            return u''
+
+    def copy(self):
+        return self.__class__(self)
 
     def __repr__(self):
         return '%s(%s)' % (
@@ -115,57 +136,6 @@ class _ExceptionCatcher(object):
         return True
 
 
-class StackedObject(object):
-    """Baseclass for all objects that provide stack manipulation
-    operations.
-    """
-
-    def push_thread(self):
-        """Pushes the stacked object to the thread stack."""
-        raise NotImplementedError()
-
-    def pop_thread(self):
-        """Pops the stacked object from the thread stack."""
-        raise NotImplementedError()
-
-    def push_application(self):
-        """Pushes the stacked object to the application stack."""
-        raise NotImplementedError()
-
-    def pop_application(self):
-        """Pops the stacked object from the application stack."""
-        raise NotImplementedError()
-
-    @contextmanager
-    def threadbound(self):
-        """Can be used in combination with the `with` statement to
-        execute code while the object is bound to the thread.
-        """
-        self.push_thread()
-        try:
-            yield self
-        finally:
-            self.pop_thread()
-
-    @contextmanager
-    def applicationbound(self):
-        """Can be used in combination with the `with` statement to
-        execute code while the object is bound to the application.
-        """
-        self.push_application()
-        try:
-            yield self
-        finally:
-            self.pop_application()
-
-    def __enter__(self):
-        self.push_thread()
-        return self
-
-    def __exit__(self, exc_type, exc_value, tb):
-        self.pop_thread()
-
-
 class ContextObject(StackedObject):
     """An object that can be bound to a context.  It is managed by the
     :class:`ContextStackManager`"""
@@ -246,8 +216,10 @@ class Processor(ContextObject):
 
 class _InheritedType(object):
     __slots__ = ()
+
     def __repr__(self):
         return 'Inherit'
+
     def __reduce__(self):
         return 'Inherit'
 Inherit = _InheritedType()
@@ -469,14 +441,17 @@ class LogRecord(object):
             # information was not pulled and the log record no longer has
             # access to the frame.  But there is not much we can do about
             # that.
-            raise TypeError(F('Could not format message with provided '
-                              'arguments: {err}\n  msg=\'{msg}\'\n  '
-                              'args={args} \n  kwargs={kwargs}.\n'
-                              'Happened in file {file}, line {lineno}').format(
-                err=e, msg=self.msg.encode('utf-8'), args=self.args,
-                kwargs=self.kwargs, file=self.filename.encode('utf-8'),
+            errormsg = F('Could not format message with provided '
+                         'arguments: {err}\n  msg=\'{msg}\'\n  '
+                         'args={args} \n  kwargs={kwargs}.\n'
+                         'Happened in file {file}, line {lineno}').format(
+                err=e, msg=self.msg, args=self.args,
+                kwargs=self.kwargs, file=self.filename,
                 lineno=self.lineno
-            ))
+            )
+            if not _py3:
+                errormsg = errormsg.encode('utf-8')
+            raise TypeError(errormsg)
 
     level_name = level_name_property()
 
@@ -521,8 +496,7 @@ class LogRecord(object):
             fn = cf.f_code.co_filename
             if fn[:1] == '<' and fn[-1:] == '>':
                 return fn
-            return os.path.abspath(fn).decode(sys.getfilesystemencoding()
-                                              or 'utf-8', 'replace')
+            return _convert_frame_filename(os.path.abspath(fn))
 
     @cached_property
     def lineno(self):
@@ -569,8 +543,9 @@ class LogRecord(object):
         in case there was any.
         """
         if self.exc_info is not None:
-            lines = traceback.format_exception(*self.exc_info)
-            rv = ''.join(lines).decode('utf-8', 'replace')
+            rv = ''.join(traceback.format_exception(*self.exc_info))
+            if not _py3:
+                rv = rv.decode('utf-8', 'replace')
             return rv.rstrip()
 
     @cached_property
@@ -623,14 +598,14 @@ class LoggerMixin(object):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.DEBUG`.
         """
-        if DEBUG >= self.level:
+        if not self.disabled and DEBUG >= self.level:
             self._log(DEBUG, args, kwargs)
 
     def info(self, *args, **kwargs):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.INFO`.
         """
-        if INFO >= self.level:
+        if not self.disabled and INFO >= self.level:
             self._log(INFO, args, kwargs)
 
     def warn(self, *args, **kwargs):
@@ -638,31 +613,35 @@ class LoggerMixin(object):
         to :data:`~logbook.WARNING`.  This function has an alias
         named :meth:`warning`.
         """
-        if WARNING >= self.level:
+        if not self.disabled and WARNING >= self.level:
             self._log(WARNING, args, kwargs)
 
     def warning(self, *args, **kwargs):
-        """ALias for :meth:`warn`."""
+        """Alias for :meth:`warn`."""
         return self.warn(*args, **kwargs)
 
     def notice(self, *args, **kwargs):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.NOTICE`.
         """
-        if NOTICE >= self.level:
+        if not self.disabled and NOTICE >= self.level:
             self._log(NOTICE, args, kwargs)
 
     def error(self, *args, **kwargs):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.ERROR`.
         """
-        if ERROR >= self.level:
+        if not self.disabled and ERROR >= self.level:
             self._log(ERROR, args, kwargs)
 
     def exception(self, *args, **kwargs):
         """Works exactly like :meth:`error` just that the message
         is optional and exception information is recorded.
         """
+        if self.disabled or ERROR < self.level:
+            return
+        if not args:
+            args = ('Uncaught exception occurred',)
         if 'exc_info' not in kwargs:
             exc_info = sys.exc_info()
             assert exc_info[0] is not None, 'no exception occurred'
@@ -673,7 +652,7 @@ class LoggerMixin(object):
         """Logs a :class:`~logbook.LogRecord` with the level set
         to :data:`~logbook.CRITICAL`.
         """
-        if CRITICAL >= self.level:
+        if not self.disabled and CRITICAL >= self.level:
             self._log(CRITICAL, args, kwargs)
 
     def log(self, level, *args, **kwargs):
@@ -738,7 +717,8 @@ class RecordDispatcher(object):
         if not self.disabled and record.level >= self.level:
             self.call_handlers(record)
 
-    def make_record_and_handle(self, level, msg, args, kwargs, exc_info, extra):
+    def make_record_and_handle(self, level, msg, args, kwargs, exc_info,
+                               extra):
         """Creates a record from some given arguments and heads it
         over to the handling system.
         """
@@ -787,7 +767,8 @@ class RecordDispatcher(object):
         # Both logger attached handlers as well as context specific
         # handlers are handled one after another.  The latter also
         # include global handlers.
-        for handler in chain(self.handlers, Handler.stack_manager.iter_context_objects()):
+        for handler in chain(self.handlers,
+                             Handler.stack_manager.iter_context_objects()):
             # skip records that this handler is not interested in based
             # on the record and handler level or in case this method was
             # overridden on some custom logic.
