@@ -8,25 +8,26 @@
     :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
     :license: BSD, see LICENSE for more details.
 """
-from __future__ import with_statement
-
 import os
 import sys
 import stat
 import errno
-import codecs
 import socket
-import hashlib
+try:
+    from hashlib import sha1
+except ImportError:
+    from sha import new as sha1
 import threading
 import traceback
 from datetime import datetime, timedelta
 from itertools import izip
 from threading import Lock
+from collections import deque
 
 from logbook.base import CRITICAL, ERROR, WARNING, NOTICE, INFO, DEBUG, \
      NOTSET, level_name_property, _missing, lookup_level, \
      Flags, ContextObject, ContextStackManager
-from logbook.helpers import rename, F
+from logbook.helpers import any, rename, F, b, _is_text_stream
 
 
 DEFAULT_FORMAT_STRING = (
@@ -58,11 +59,20 @@ Time:               {record.time:%Y-%m-%d %H:%M:%S}
 
 Message:
 
+{record.message}
+'''
+MAIL_RELATED_FORMAT_STRING = u'''\
+Message type:       {record.level_name}
+Location:           {record.filename}:{record.lineno}
+Module:             {record.module}
+Function:           {record.func_name}
 {record.message}
 '''
 
 SYSLOG_PORT = 514
 
+_py3 = sys.version_info >= (3, 0)
+
 
 def create_syshandler(application_name, level=NOTSET):
     """Creates the handler the operating system provides.  On Unix systems
@@ -212,6 +222,59 @@ class Handler(ContextObject):
         log records.
         """
 
+    def emit_batch(self, records, reason):
+        """Some handlers may internally queue up records and want to forward
+        them at once to another handler.  For example the
+        :class:`~logbook.FingersCrossedHandler` internally buffers
+        records until a level threshold is reached in which case the buffer
+        is sent to this method and not :meth:`emit` for each record.
+
+        The default behaviour is to call :meth:`emit` for each record in
+        the buffer, but handlers can use this to optimize log handling.  For
+        instance the mail handler will try to batch up items into one mail
+        and not to emit mails for each record in the buffer.
+
+        Note that unlike :meth:`emit` there is no wrapper method like
+        :meth:`handle` that does error handling.  The reason is that this
+        is intended to be used by other handlers which are already protected
+        against internal breakage.
+
+        `reason` is a string that specifies the rason why :meth:`emit_batch`
+        was called, and not :meth:`emit`.  The following are valid values:
+
+        ``'buffer'``
+            Records were buffered for performance reasons or because the
+            records were sent to another process and buffering was the only
+            possible way.  For most handlers this should be equivalent to
+            calling :meth:`emit` for each record.
+
+        ``'escalation'``
+            Escalation means that records were buffered in case the threshold
+            was exceeded.  In this case, the last record in the iterable is the
+            record that triggered the call.
+
+        ``'group'``
+            All the records in the iterable belong to the same logical
+            component and happened in the same process.  For example there was
+            a long running computation and the handler is invoked with a bunch
+            of records that happened there.  This is similar to the escalation
+            reason, just that the first one is the significant one, not the
+            last.
+
+        If a subclass overrides this and does not want to handle a specific
+        reason it must call into the superclass because more reasons might
+        appear in future releases.
+
+        Example implementation::
+
+            def emit_batch(self, records, reason):
+                if reason not in ('escalation', 'group'):
+                    Handler.emit_batch(self, records, reason)
+                ...
+        """
+        for record in records:
+            self.emit(record)
+
     def close(self):
         """Tidy up any resources used by the handler.  This is automatically
         called by the destructor of the class as well, but explicit calls are
@@ -282,9 +345,11 @@ class StringFormatter(object):
 
     def _get_format_string(self):
         return self._format_string
+
     def _set_format_string(self, value):
         self._format_string = value
         self._formatter = F(value)
+
     format_string = property(_get_format_string, _set_format_string)
     del _get_format_string, _set_format_string
 
@@ -326,11 +391,13 @@ class StringFormatterHandlerMixin(object):
     def _get_format_string(self):
         if isinstance(self.formatter, StringFormatter):
             return self.formatter.format_string
+
     def _set_format_string(self, value):
         if value is None:
             self.formatter = None
         else:
             self.formatter = self.formatter_class(value)
+
     format_string = property(_get_format_string, _set_format_string)
     del _get_format_string, _set_format_string
 
@@ -340,11 +407,11 @@ class HashingHandlerMixin(object):
 
     def hash_record_raw(self, record):
         """Returns a hashlib object with the hash of the record."""
-        hash = hashlib.sha1()
-        hash.update('%d\x00' % record.level)
-        hash.update((record.channel or u'').encode('utf-8') + '\x00')
-        hash.update(record.filename.encode('utf-8') + '\x00')
-        hash.update(str(record.lineno))
+        hash = sha1()
+        hash.update(('%d\x00' % record.level).encode('ascii'))
+        hash.update((record.channel or u'').encode('utf-8') + b('\x00'))
+        hash.update(record.filename.encode('utf-8') + b('\x00'))
+        hash.update(b(str(record.lineno)))
         return hash
 
     def hash_record(self, record):
@@ -382,15 +449,15 @@ class LimitingHandlerMixin(HashingHandlerMixin):
 
     def check_delivery(self, record):
         """Helper function to check if data should be delivered by this
-        handler.  It returns a tuple in the form ``(suppression_count, allow)``.
-        The first one is the number of items that were not delivered so far,
-        the second is a boolean flag if a delivery should happen now.
+        handler.  It returns a tuple in the form ``(suppression_count,
+        allow)``.  The first one is the number of items that were not delivered
+        so far, the second is a boolean flag if a delivery should happen now.
         """
         if self.record_limit is None:
             return 0, True
         hash = self.hash_record(record)
-        with self._limit_lock:
-
+        self._limit_lock.acquire()
+        try:
             allow_delivery = None
             suppression_count = old_count = 0
             first_count = now = datetime.utcnow()
@@ -407,7 +474,8 @@ class LimitingHandlerMixin(HashingHandlerMixin):
                len(self._record_limits) >= self.max_record_cache:
                 cache_items = self._record_limits.items()
                 cache_items.sort()
-                del cache_items[:int(self._record_limits) * self.record_cache_prune]
+                del cache_items[:int(self._record_limits) \
+                    * self.record_cache_prune]
                 self._record_limits = dict(cache_items)
 
             self._record_limits[hash] = (first_count, old_count + 1)
@@ -415,6 +483,8 @@ class LimitingHandlerMixin(HashingHandlerMixin):
             if allow_delivery is None:
                 allow_delivery = old_count < self.record_limit
             return suppression_count, allow_delivery
+        finally:
+            self._limit_lock.release()
 
 
 class StreamHandler(Handler, StringFormatterHandlerMixin):
@@ -427,12 +497,18 @@ class StreamHandler(Handler, StringFormatterHandlerMixin):
 
         with StreamHandler(my_stream):
             pass
+
+    .. admonition:: Notes on the encoding
+
+       On Python 3, the encoding parameter is only used if a stream was
+       passed that was opened in binary mode.
     """
 
-    def __init__(self, stream, level=NOTSET, format_string=None, filter=None,
-                 bubble=False):
+    def __init__(self, stream, level=NOTSET, format_string=None,
+                 encoding=None, filter=None, bubble=False):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
+        self.encoding = encoding
         self.lock = threading.Lock()
         if stream is not _missing:
             self.stream = stream
@@ -457,17 +533,26 @@ class StreamHandler(Handler, StringFormatterHandlerMixin):
 
     def format_and_encode(self, record):
         """Formats the record and encodes it to the stream encoding."""
-        enc = getattr(self.stream, 'encoding', None) or 'utf-8'
-        return (self.format(record) + u'\n').encode(enc, 'replace')
+        stream = self.stream
+        rv = self.format(record) + '\n'
+        if not _py3 or not _is_text_stream(stream):
+            enc = self.encoding
+            if enc is None:
+                enc = getattr(stream, 'encoding', None) or 'utf-8'
+            rv = rv.encode(enc, 'replace')
+        return rv
 
     def write(self, item):
         """Writes a bytestring to the stream."""
         self.stream.write(item)
 
     def emit(self, record):
-        with self.lock:
+        self.lock.acquire()
+        try:
             self.write(self.format_and_encode(record))
             self.flush()
+        finally:
+            self.lock.release()
 
 
 class FileHandler(StreamHandler):
@@ -476,15 +561,17 @@ class FileHandler(StreamHandler):
     the open to the point where the first message is written.
 
     This is useful when the handler is used with a
-    :class:`~logbook.more.FingersCrossedHandler` or something similar.
+    :class:`~logbook.FingersCrossedHandler` or something similar.
     """
 
-    def __init__(self, filename, mode='a', encoding='utf-8', level=NOTSET,
+    def __init__(self, filename, mode='a', encoding=None, level=NOTSET,
                  format_string=None, delay=False, filter=None, bubble=False):
-        StreamHandler.__init__(self, None, level, format_string, filter, bubble)
+        if encoding is None:
+            encoding = 'utf-8'
+        StreamHandler.__init__(self, None, level, format_string,
+                               encoding, filter, bubble)
         self._filename = filename
         self._mode = mode
-        self._encoding = encoding
         if delay:
             self.stream = None
         else:
@@ -493,12 +580,15 @@ class FileHandler(StreamHandler):
     def _open(self, mode=None):
         if mode is None:
             mode = self._mode
-        self.stream = codecs.open(self._filename, mode, self._encoding)
+        self.stream = open(self._filename, mode)
 
     def write(self, item):
         if self.stream is None:
             self._open()
-        StreamHandler.write(self, item)
+        if _py3 and isinstance(item, bytes):
+            self.stream.buffer.write(item)
+        else:
+            self.stream.write(item)
 
     def close(self):
         if self.stream is not None:
@@ -506,6 +596,13 @@ class FileHandler(StreamHandler):
             self.stream.close()
             self.stream = None
 
+    def format_and_encode(self, record):
+        # encodes based on the stream settings, so the stream has to be
+        # open at the time this function is called.
+        if self.stream is None:
+            self._open()
+        return StreamHandler.format_and_encode(self, record)
+
     def emit(self, record):
         if self.stream is None:
             self._open()
@@ -526,7 +623,8 @@ class MonitoringFileHandler(FileHandler):
         FileHandler.__init__(self, filename, mode, encoding, level,
                              format_string, delay, filter, bubble)
         if os.name == 'nt':
-            raise RuntimeError('MonitoringFileHandler does not support Windows')
+            raise RuntimeError('MonitoringFileHandler '
+                               'does not support Windows')
         self._query_fd()
 
     def _query_fd(self):
@@ -563,7 +661,7 @@ class StderrHandler(StreamHandler):
     def __init__(self, level=NOTSET, format_string=None, filter=None,
                  bubble=False):
         StreamHandler.__init__(self, _missing, level, format_string,
-                               filter, bubble)
+                               None, filter, bubble)
 
     @property
     def stream(self):
@@ -571,21 +669,39 @@ class StderrHandler(StreamHandler):
 
 
 class RotatingFileHandlerBase(FileHandler):
-    """Baseclass for rotating file handlers."""
+    """Baseclass for rotating file handlers.
+
+    .. versionchanged:: 0.3
+       This class was deprecated because the interface is not flexible
+       enough to implement proper file rotations.  The former builtin
+       subclasses no longer use this baseclass.
+    """
+
+    def __init__(self, *args, **kwargs):
+        from warnings import warn
+        warn(DeprecationWarning('This class is deprecated'))
+        FileHandler.__init__(self, *args, **kwargs)
 
     def emit(self, record):
-        with self.lock:
+        self.lock.acquire()
+        try:
             msg = self.format_and_encode(record)
-            if self.should_rollover(record, len(msg)):
+            if self.should_rollover(record, msg):
                 self.perform_rollover()
             self.write(msg)
             self.flush()
+        finally:
+            self.lock.release()
 
-    def should_rollover(self, record, bytes):
-        """Called with the log record and the number of bytes that
-        would be written into the file.  The method has then to
+    def should_rollover(self, record, formatted_record):
+        """Called with the log record and the return value of the
+        :meth:`format_and_encode` method.  The method has then to
         return `True` if a rollover should happen or `False`
         otherwise.
+
+        .. versionchanged:: 0.3
+           Previously this method was called with the number of bytes
+           returned by :meth:`format_and_encode`
         """
         return False
 
@@ -595,7 +711,7 @@ class RotatingFileHandlerBase(FileHandler):
         """
 
 
-class RotatingFileHandler(RotatingFileHandlerBase):
+class RotatingFileHandler(FileHandler):
     """This handler rotates based on file size.  Once the maximum size
     is reached it will reopen the file and start with an empty file
     again.  The old file is moved into a backup copy (named like the
@@ -612,8 +728,8 @@ class RotatingFileHandler(RotatingFileHandlerBase):
     def __init__(self, filename, mode='a', encoding='utf-8', level=NOTSET,
                  format_string=None, delay=False, max_size=1024 * 1024,
                  backup_count=5, filter=None, bubble=False):
-        RotatingFileHandlerBase.__init__(self, filename, mode, encoding, level,
-                                         format_string, delay, filter, bubble)
+        FileHandler.__init__(self, filename, mode, encoding, level,
+                             format_string, delay, filter, bubble)
         self.max_size = max_size
         self.backup_count = backup_count
         assert backup_count > 0, 'at least one backup file has to be ' \
@@ -636,8 +752,19 @@ class RotatingFileHandler(RotatingFileHandlerBase):
         rename(self._filename, self._filename + '.1')
         self._open('w')
 
+    def emit(self, record):
+        self.lock.acquire()
+        try:
+            msg = self.format_and_encode(record)
+            if self.should_rollover(record, len(msg)):
+                self.perform_rollover()
+            self.write(msg)
+            self.flush()
+        finally:
+            self.lock.release()
 
-class TimedRotatingFileHandler(RotatingFileHandlerBase):
+
+class TimedRotatingFileHandler(FileHandler):
     """This handler rotates based on dates.  It will name the file
     after the filename you specify and the `date_format` pattern.
 
@@ -659,8 +786,8 @@ class TimedRotatingFileHandler(RotatingFileHandlerBase):
     def __init__(self, filename, mode='a', encoding='utf-8', level=NOTSET,
                  format_string=None, date_format='%Y-%m-%d',
                  backup_count=0, filter=None, bubble=False):
-        RotatingFileHandlerBase.__init__(self, filename, mode, encoding, level,
-                                         format_string, True, filter, bubble)
+        FileHandler.__init__(self, filename, mode, encoding, level,
+                             format_string, True, filter, bubble)
         self.date_format = date_format
         self.backup_count = backup_count
         self._fn_parts = os.path.splitext(os.path.abspath(filename))
@@ -670,7 +797,7 @@ class TimedRotatingFileHandler(RotatingFileHandlerBase):
         return datetime.strftime('-' + self.date_format) \
                        .join(self._fn_parts)
 
-    def should_rollover(self, record, bytes):
+    def should_rollover(self, record):
         fn = self._get_timed_filename(record.time)
         rv = self._filename is not None and self._filename != fn
         # remember the current filename.  In case rv is True, the rollover
@@ -699,6 +826,16 @@ class TimedRotatingFileHandler(RotatingFileHandlerBase):
                 os.remove(filename)
         self._open('w')
 
+    def emit(self, record):
+        self.lock.acquire()
+        try:
+            if self.should_rollover(record):
+                self.perform_rollover()
+            self.write(self.format_and_encode(record))
+            self.flush()
+        finally:
+            self.lock.release()
+
 
 class TestHandler(Handler, StringFormatterHandlerMixin):
     """Like a stream handler but keeps the values in memory.  This
@@ -714,7 +851,8 @@ class TestHandler(Handler, StringFormatterHandlerMixin):
     """
     default_format_string = TEST_FORMAT_STRING
 
-    def __init__(self, level=NOTSET, format_string=None, filter=None, bubble=False):
+    def __init__(self, level=NOTSET, format_string=None, filter=None,
+                 bubble=False):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
         #: captures the :class:`LogRecord`\s as instances
@@ -865,8 +1003,12 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
     of ZeroMQ or others) and the logging system slows you down you can
     wrap the handler in a :class:`logbook.queues.ThreadedWrapperHandler`
     that will then send the mails in a background thread.
+
+    .. versionchanged:: 0.3
+       The handler supports the batching system now.
     """
     default_format_string = MAIL_FORMAT_STRING
+    default_related_format_string = MAIL_RELATED_FORMAT_STRING
     default_subject = u'Server Error in Application'
 
     #: the maximum number of record hashes in the cache for the limiting
@@ -880,7 +1022,8 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
     def __init__(self, from_addr, recipients, subject=None,
                  server_addr=None, credentials=None, secure=None,
                  record_limit=None, record_delta=None, level=NOTSET,
-                 format_string=None, filter=None, bubble=False):
+                 format_string=None, related_format_string=None,
+                 filter=None, bubble=False):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
         LimitingHandlerMixin.__init__(self, record_limit, record_delta)
@@ -892,6 +1035,21 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         self.server_addr = server_addr
         self.credentials = credentials
         self.secure = secure
+        if related_format_string is None:
+            related_format_string = self.default_related_format_string
+        self.related_format_string = related_format_string
+
+    def _get_related_format_string(self):
+        if isinstance(self.related_formatter, StringFormatter):
+            return self.related_formatter.format_string
+    def _set_related_format_string(self, value):
+        if value is None:
+            self.related_formatter = None
+        else:
+            self.related_formatter = self.formatter_class(value)
+    related_format_string = property(_get_related_format_string,
+                                    _set_related_format_string)
+    del _get_related_format_string, _set_related_format_string
 
     def get_recipients(self, record):
         """Returns the recipients for a record.  By default the
@@ -904,7 +1062,10 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         (:class:`email.message.Message`).  `suppressed` is the number
         of mails not sent if the `record_limit` feature is active.
         """
-        from email.message import Message
+        try:
+            from email.message import Message
+        except ImportError:  # Python 2.4
+            from email.Message import Message
         msg = Message()
         lineiter = iter(self.format(record).splitlines())
         for line in lineiter:
@@ -919,17 +1080,41 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         msg.set_payload(body)
         return msg
 
+    def format_related_record(self, record):
+        """Used for format the records that led up to another record or
+        records that are related into strings.  Used by the batch formatter.
+        """
+        return self.related_formatter(record, self)
+
     def generate_mail(self, record, suppressed=0):
         """Generates the final email (:class:`email.message.Message`)
         with headers and date.  `suppressed` is the number of mails
         that were not send if the `record_limit` feature is active.
         """
-        from email.utils import formatdate
+        try:
+            from email.utils import formatdate
+        except ImportError:  # Python 2.4
+            from email.Utils import formatdate
         msg = self.message_from_record(record, suppressed)
         msg['From'] = self.from_addr
         msg['Date'] = formatdate()
         return msg
 
+    def collapse_mails(self, mail, related, reason):
+        """When escaling or grouped mails are """
+        if not related:
+            return mail
+        if reason == 'group':
+            title = 'Other log records in the same group'
+        else:
+            title = 'Log records that led up to this one'
+        mail.set_payload('%s\r\n\r\n\r\n%s:\r\n\r\n%s' % (
+            mail.get_payload(),
+            title,
+            '\r\n\r\n'.join(body.rstrip() for body in related)
+        ))
+        return mail
+
     def get_connection(self):
         """Returns an SMTP connection.  By default it reconnects for
         each sent mail.
@@ -937,7 +1122,7 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         from smtplib import SMTP, SMTP_PORT, SMTP_SSL_PORT
         if self.server_addr is None:
             host = ''
-            port = SMTP_SSL_PORT if self.secure else SMTP_PORT
+            port = self.secure and SMTP_SSL_PORT or SMTP_PORT
         else:
             host, port = self.server_addr
         con = SMTP()
@@ -977,6 +1162,27 @@ class MailHandler(Handler, StringFormatterHandlerMixin,
         self.deliver(self.generate_mail(record, suppressed),
                      self.get_recipients(record))
 
+    def emit_batch(self, records, reason):
+        if reason not in ('escalation', 'group'):
+            return MailHandler.emit_batch(self, records, reason)
+        records = list(records)
+        if not records:
+            return
+
+        trigger = records.pop(reason == 'escalation' and -1 or 0)
+        suppressed = 0
+        if self.record_limit is not None:
+            suppressed, allow_delivery = self.check_delivery(trigger)
+            if not allow_delivery:
+                return
+
+        trigger_mail = self.generate_mail(trigger, suppressed)
+        related = [self.format_related_record(record)
+                   for record in records]
+
+        self.deliver(self.collapse_mails(trigger_mail, related, reason),
+                     self.get_recipients(trigger))
+
 
 class SyslogHandler(Handler, StringFormatterHandlerMixin):
     """A handler class which sends formatted logging records to a
@@ -1097,12 +1303,14 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
         return (facility << 3) | priority
 
     def emit(self, record):
-        prefix = ''
+        prefix = u''
         if self.application_name is not None:
-            prefix = self.application_name.encode('utf-8') + ':'
-        message = self.format(record).encode('utf-8')
-        self.send_to_socket('<%d>%s%s\x00' % (self.encode_priority(record),
-                                              prefix, message))
+            prefix = self.application_name + u':'
+        self.send_to_socket((u'<%d>%s%s\x00' % (
+            self.encode_priority(record),
+            prefix,
+            self.format(record)
+        )).encode('utf-8'))
 
     def send_to_socket(self, data):
         if self.unixsocket:
@@ -1112,7 +1320,8 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
                 self._connect_unixsocket()
                 self.socket.send(data)
         elif self.socktype == socket.SOCK_DGRAM:
-            self.socket.sendto(data, self.address)
+            # the flags are no longer optional on Python 3
+            self.socket.sendto(data, 0, self.address)
         else:
             self.socket.sendall(data)
 
@@ -1136,7 +1345,8 @@ class NTEventLogHandler(Handler, StringFormatterHandlerMixin):
                                'operating system.')
 
         try:
-            import win32evtlogutil, win32evtlog
+            import win32evtlogutil
+            import win32evtlog
         except ImportError:
             raise RuntimeError('The pywin32 library is required '
                                'for the NTEventLogHandler.')
@@ -1184,3 +1394,189 @@ class NTEventLogHandler(Handler, StringFormatterHandlerMixin):
         type = self.get_event_type(record)
         self._welu.ReportEvent(self.application_name, id, cat, type,
                                [self.format(record)])
+
+
+class FingersCrossedHandler(Handler):
+    """This handler wraps another handler and will log everything in
+    memory until a certain level (`action_level`, defaults to `ERROR`)
+    is exceeded.  When that happens the fingers crossed handler will
+    activate forever and log all buffered records as well as records
+    yet to come into another handled which was passed to the constructor.
+
+    Alternatively it's also possible to pass a factory function to the
+    constructor instead of a handler.  That factory is then called with
+    the triggering log entry and the finger crossed handler to create
+    a handler which is then cached.
+
+    The idea of this handler is to enable debugging of live systems.  For
+    example it might happen that code works perfectly fine 99% of the time,
+    but then some exception happens.  But the error that caused the
+    exception alone might not be the interesting bit, the interesting
+    information were the warnings that lead to the error.
+
+    Here a setup that enables this for a web application::
+
+        from logbook import FileHandler
+        from logbook import FingersCrossedHandler
+
+        def issue_logging():
+            def factory(record, handler):
+                return FileHandler('/var/log/app/issue-%s.log' % record.time)
+            return FingersCrossedHandler(factory)
+
+        def application(environ, start_response):
+            with issue_logging():
+                return the_actual_wsgi_application(environ, start_response)
+
+    Whenever an error occours, a new file in ``/var/log/app`` is created
+    with all the logging calls that lead up to the error up to the point
+    where the `with` block is exited.
+
+    Please keep in mind that the :class:`~logbook.FingersCrossedHandler`
+    handler is a one-time handler.  Once triggered, it will not reset.  Because
+    of that you will have to re-create it whenever you bind it.  In this case
+    the handler is created when it's bound to the thread.
+
+    Due to how the handler is implemented, the filter, bubble and level
+    flags of the wrapped handler are ignored.
+
+    .. versionchanged:: 0.3
+
+    The default behaviour is to buffer up records and then invoke another
+    handler when a severity theshold was reached with the buffer emitting.
+    This now enables this logger to be properly used with the
+    :class:`~logbook.MailHandler`.  You will now only get one mail for
+    each bfufered record.  However once the threshold was reached you would
+    still get a mail for each record which is why the `reset` flag was added.
+
+    When set to `True`, the handler will instantly reset to the untriggered
+    state and start buffering again::
+
+        handler = FingersCrossedHandler(MailHandler(...),
+                                        buffer_size=10,
+                                        reset=True)
+
+    .. versionadded:: 0.3
+       The `reset` flag was added.
+    """
+
+    #: the reason to be used for the batch emit.  The default is
+    #: ``'escalation'``.
+    #:
+    #: .. versionadded:: 0.3
+    batch_emit_reason = 'escalation'
+
+    def __init__(self, handler, action_level=ERROR, buffer_size=0,
+                 pull_information=True, reset=False, filter=None,
+                 bubble=False):
+        Handler.__init__(self, NOTSET, filter, bubble)
+        self.lock = Lock()
+        self._level = action_level
+        if isinstance(handler, Handler):
+            self._handler = handler
+            self._handler_factory = None
+        else:
+            self._handler = None
+            self._handler_factory = handler
+        #: the buffered records of the handler.  Once the action is triggered
+        #: (:attr:`triggered`) this list will be None.  This attribute can
+        #: be helpful for the handler factory function to select a proper
+        #: filename (for example time of first log record)
+        self.buffered_records = deque()
+        #: the maximum number of entries in the buffer.  If this is exhausted
+        #: the oldest entries will be discarded to make place for new ones
+        self.buffer_size = buffer_size
+        self._buffer_full = False
+        self._pull_information = pull_information
+        self._action_triggered = False
+        self._reset = reset
+
+    def close(self):
+        if self._handler is not None:
+            self._handler.close()
+
+    def enqueue(self, record):
+        if self._pull_information:
+            record.pull_information()
+        if self._action_triggered:
+            self._handler.emit(record)
+        else:
+            self.buffered_records.append(record)
+            if self._buffer_full:
+                self.buffered_records.popleft()
+            elif self.buffer_size and \
+                 len(self.buffered_records) >= self.buffer_size:
+                self._buffer_full = True
+            return record.level >= self._level
+        return False
+
+    def rollover(self, record):
+        if self._handler is None:
+            self._handler = self._handler_factory(record, self)
+        self._handler.emit_batch(iter(self.buffered_records), 'escalation')
+        self.buffered_records.clear()
+        self._action_triggered = not self._reset
+
+    @property
+    def triggered(self):
+        """This attribute is `True` when the action was triggered.  From
+        this point onwards the finger crossed handler transparently
+        forwards all log records to the inner handler.  If the handler resets
+        itself this will always be `False`.
+        """
+        return self._action_triggered
+
+    def emit(self, record):
+        self.lock.acquire()
+        try:
+            if self.enqueue(record):
+                self.rollover(record)
+        finally:
+            self.lock.release()
+
+
+class GroupHandler(WrapperHandler):
+    """A handler that buffers all messages until it is popped again and then
+    forwards all messages to another handler.  This is useful if you for
+    example have an application that does computations and only a result
+    mail is required.  A group handler makes sure that only one mail is sent
+    and not multiple.  Some other handles might support this as well, though
+    currently none of the builtins do.
+
+    Example::
+
+        with GroupHandler(MailHandler(...)):
+            # everything here ends up in the mail
+
+    The :class:`GroupHandler` is implemented as a :class:`WrapperHandler`
+    thus forwarding all attributes of the wrapper handler.
+
+    Notice that this handler really only emit the records when the handler
+    is popped from the stack.
+
+    .. versionadded:: 0.3
+    """
+    _direct_attrs = frozenset(['handler', 'pull_information',
+                               'buffered_records'])
+
+    def __init__(self, handler, pull_information=True):
+        WrapperHandler.__init__(self, handler)
+        self.pull_information = pull_information
+        self.buffered_records = []
+
+    def rollover(self):
+        self.handler.emit_batch(self.buffered_records, 'group')
+        self.buffered_records = []
+
+    def pop_application(self):
+        Handler.pop_application(self)
+        self.rollover()
+
+    def pop_thread(self):
+        Handler.pop_thread(self)
+        self.rollover()
+
+    def emit(self, record):
+        if self.pull_information:
+            record.pull_information()
+        self.buffered_records.append(record)
