@@ -17,6 +17,17 @@ import random
 from datetime import datetime, timedelta
 
 
+# Python 2.4 compatibility
+
+try:
+    any = any
+except NameError:
+    def any(iterable):
+        for item in iterable:
+            if item:
+                return True
+
+
 # Python 2.5 compatibility
 
 try:
@@ -39,6 +50,14 @@ _iso8601_re = re.compile(
     r'(?:T(\d{2}):(\d{2})(?::(\d{2}(?:\.\d+)?))?(Z|[+-]\d{2}:\d{2})?)?$'
 )
 _missing = object()
+_py3 = sys.version_info >= (3, 0)
+if _py3:
+    import io
+    def b(x): return x.encode('ascii')
+    def _is_text_stream(stream): return isinstance(stream, io.TextIOBase)
+else:
+    def b(x): return x
+    def _is_text_stream(x): return True
 
 
 can_rename_open_file = False
@@ -129,7 +148,7 @@ def to_safe_json(data):
     def _convert(obj):
         if obj is None:
             return None
-        elif isinstance(obj, str):
+        elif not _py3 and isinstance(obj, str):
             return obj.decode('utf-8', 'replace')
         elif isinstance(obj, (bool, int, long, float, unicode)):
             return obj
@@ -142,7 +161,7 @@ def to_safe_json(data):
         elif isinstance(obj, dict):
             rv = {}
             for key, value in obj.iteritems():
-                if isinstance(key, str):
+                if not _py3 and isinstance(key, str):
                     key = key.decode('utf-8', 'replace')
                 else:
                     key = unicode(key)
