@@ -8,21 +8,16 @@
     :copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
     :license: BSD, see LICENSE for more details.
 """
-from __future__ import with_statement
-
 import re
 import os
-import sys
-import time
-from collections import deque
-from threading import Lock
 from cgi import parse_qsl
 from urllib import urlencode
 
-
-from logbook.base import LogRecord, RecordDispatcher, NOTSET, ERROR, WARNING
-from logbook.handlers import Handler, StringFormatter, StringFormatterHandlerMixin
-from logbook.helpers import json, get_application_name
+from logbook.base import RecordDispatcher, NOTSET, ERROR, NOTICE
+from logbook.handlers import Handler, StringFormatter, \
+     StringFormatterHandlerMixin, StderrHandler
+from logbook._termcolors import colorize
+from logbook.helpers import F
 
 
 _ws_re = re.compile(r'(\s+)(?u)')
@@ -32,6 +27,30 @@ TWITTER_ACCESS_TOKEN_URL = 'https://twitter.com/oauth/access_token'
 NEW_TWEET_URL = 'https://api.twitter.com/1/statuses/update.json'
 
 
+class TwitterFormatter(StringFormatter):
+    """Works like the standard string formatter and is used by the
+    :class:`TwitterHandler` unless changed.
+    """
+    max_length = 140
+
+    def format_exception(self, record):
+        return u'%s: %s' % (record.exception_shortname,
+                            record.exception_message)
+
+    def __call__(self, record, handler):
+        formatted = StringFormatter.__call__(self, record, handler)
+        rv = []
+        length = 0
+        for piece in _ws_re.split(formatted):
+            length += len(piece)
+            if length > self.max_length:
+                if length - len(piece) < self.max_length:
+                    rv.append(u'…')
+                break
+            rv.append(piece)
+        return u''.join(rv)
+
+
 class TaggingLogger(RecordDispatcher):
     """A logger that attaches a tag to each record.  This is an alternative
     record dispatcher that does not use levels but tags to keep log
@@ -64,7 +83,8 @@ class TaggingLogger(RecordDispatcher):
         exc_info = kwargs.pop('exc_info', None)
         extra = kwargs.pop('extra', {})
         extra['tags'] = list(tags)
-        return self.make_record_and_handle(NOTSET, msg, args, kwargs, exc_info, extra)
+        return self.make_record_and_handle(NOTSET, msg, args, kwargs,
+                                           exc_info, extra)
 
 
 class TaggingHandler(Handler):
@@ -94,138 +114,6 @@ class TaggingHandler(Handler):
                 handler.handle(record)
 
 
-class FingersCrossedHandler(Handler):
-    """This handler wraps another handler and will log everything in
-    memory until a certain level (`action_level`, defaults to `ERROR`)
-    is exceeded.  When that happens the fingers crossed handler will
-    activate forever and log all buffered records as well as records
-    yet to come into another handled which was passed to the constructor.
-
-    Alternatively it's also possible to pass a factory function to the
-    constructor instead of a handler.  That factory is then called with
-    the triggering log entry and the finger crossed handler to create
-    a handler which is then cached.
-
-    The idea of this handler is to enable debugging of live systems.  For
-    example it might happen that code works perfectly fine 99% of the time,
-    but then some exception happens.  But the error that caused the
-    exception alone might not be the interesting bit, the interesting
-    information were the warnings that lead to the error.
-
-    Here a setup that enables this for a web application::
-
-        from logbook import FileHandler
-        from logbook.more import FingersCrossedHandler
-
-        def issue_logging():
-            def factory(record, handler):
-                return FileHandler('/var/log/app/issue-%s.log' % record.time)
-            return FingersCrossedHandler(factory)
-
-        def application(environ, start_response):
-            with issue_logging():
-                return the_actual_wsgi_application(environ, start_response)
-
-    Whenever an error occours, a new file in ``/var/log/app`` is created
-    with all the logging calls that lead up to the error up to the point
-    where the `with` block is exited.
-
-    Please keep in mind that the :class:`~logbook.more.FingersCrossedHandler`
-    handler is a one-time handler.  Once triggered, it will not reset.  Because
-    of that you will have to re-create it whenever you bind it.  In this case
-    the handler is created when it's bound to the thread.
-    """
-
-    def __init__(self, handler, action_level=ERROR, buffer_size=0,
-                 pull_information=True, filter=None, bubble=False):
-        Handler.__init__(self, NOTSET, filter, bubble)
-        self.lock = Lock()
-        self._level = action_level
-        if isinstance(handler, Handler):
-            self._handler = handler
-            self._handler_factory = None
-        else:
-            self._handler = None
-            self._handler_factory = handler
-        #: the buffered records of the handler.  Once the action is triggered
-        #: (:attr:`triggered`) this list will be None.  This attribute can
-        #: be helpful for the handler factory function to select a proper
-        #: filename (for example time of first log record)
-        self.buffered_records = deque()
-        #: the maximum number of entries in the buffer.  If this is exhausted
-        #: the oldest entries will be discarded to make place for new ones
-        self.buffer_size = buffer_size
-        self._buffer_full = False
-        self._pull_information = pull_information
-        self._action_triggered = False
-
-    def close(self):
-        if self._handler is not None:
-            self._handler.close()
-
-    def enqueue(self, record):
-        assert self.buffered_records is not None, 'rollover occurred'
-        if self._pull_information:
-            record.pull_information()
-        self.buffered_records.append(record)
-        if self._buffer_full:
-            self.buffered_records.popleft()
-        elif self.buffer_size and \
-             len(self.buffered_records) >= self.buffer_size - 1:
-            self._buffer_full = True
-
-    def rollover(self, record):
-        assert self.buffered_records is not None, 'rollover occurred'
-        if self._handler is None:
-            self._handler = self._handler_factory(record, self)
-        for old_record in self.buffered_records:
-            self._handler.emit(old_record)
-        self.buffered_records = None
-        self._action_triggered = True
-
-    @property
-    def triggered(self):
-        """This attribute is `True` when the action was triggered.  From
-        this point onwards the finger crossed handler transparently
-        forwards all log records to the inner handler.
-        """
-        return self._action_triggered
-
-    def emit(self, record):
-        with self.lock:
-            if self._action_triggered:
-                self._handler.emit(record)
-            elif record.level >= self._level:
-                self.rollover(record)
-                self._handler.emit(record)
-            else:
-                self.enqueue(record)
-
-
-class TwitterFormatter(StringFormatter):
-    """Works like the standard string formatter and is used by the
-    :class:`TwitterHandler` unless changed.
-    """
-    max_length = 140
-
-    def format_exception(self, record):
-        return u'%s: %s' % (record.exception_shortname,
-                            record.exception_message)
-
-    def __call__(self, record, handler):
-        formatted = StringFormatter.__call__(self, record, handler)
-        rv = []
-        length = 0
-        for piece in _ws_re.split(formatted):
-            length += len(piece)
-            if length > self.max_length:
-                if length - len(piece) < self.max_length:
-                    rv.append(u'…')
-                break
-            rv.append(piece)
-        return u''.join(rv)
-
-
 class TwitterHandler(Handler, StringFormatterHandlerMixin):
     """A handler that logs to twitter.  Requires that you sign up an
     application on twitter and request xauth support.  Furthermore the
@@ -313,3 +201,130 @@ class JinjaFormatter(object):
 
     def __call__(self, record, handler):
         return self.template.render(record=record, handler=handler)
+
+
+class ExternalApplicationHandler(Handler):
+    """This handler invokes an external application to send parts of
+    the log record to.  The constructor takes a list of arguments that
+    are passed to another application where each of the arguments is a
+    format string, and optionally a format string for data that is
+    passed to stdin.
+
+    For example it can be used to invoke the ``say`` command on OS X::
+
+        from logbook.more import ExternalApplicationHandler
+        say_handler = ExternalApplicationHandler(['say', '{record.message}'])
+
+    Note that the above example is blocking until ``say`` finished, so it's
+    recommended to combine this handler with the
+    :class:`logbook.ThreadedWrapperHandler` to move the execution into
+    a background thread.
+
+    .. versionadded:: 0.3
+    """
+
+    def __init__(self, arguments, stdin_format=None,
+                 encoding='utf-8', level=NOTSET, filter=None,
+                 bubble=False):
+        Handler.__init__(self, level, filter, bubble)
+        self.encoding = encoding
+        self._arguments = [F(arg) for arg in arguments]
+        if stdin_format is not None:
+            stdin_format = F(stdin_format)
+        self._stdin_format = stdin_format
+        import subprocess
+        self._subprocess = subprocess
+
+    def emit(self, record):
+        args = [arg.format(record=record).encode(self.encoding)
+                for arg in self._arguments]
+        if self._stdin_format is not None:
+            stdin_data = self._stdin_format.format(record=record) \
+                                           .encode(self.encoding)
+            stdin = self._subprocess.PIPE
+        else:
+            stdin = None
+        c = self._subprocess.Popen(args, stdin=stdin)
+        if stdin is not None:
+            c.communicate(stdin_data)
+        c.wait()
+
+
+class ColorizingStreamHandlerMixin(object):
+    """A mixin class that does colorizing.
+
+    .. versionadded:: 0.3
+    """
+
+    def should_colorize(self, record):
+        """Returns `True` if colorizing should be applied to this
+        record.  The default implementation returns `True` if the
+        stream is a tty and we are not executing on windows.
+        """
+        if os.name == 'nt':
+            return False
+        isatty = getattr(self.stream, 'isatty', None)
+        return isatty and isatty()
+
+    def get_color(self, record):
+        """Returns the color for this record."""
+        if record.level >= ERROR:
+            return 'red'
+        elif record.level >= NOTICE:
+            return 'yellow'
+        return 'lightgray'
+
+    def format_and_encode(self, record):
+        rv = super(ColorizingStreamHandlerMixin, self) \
+                .format_and_encode(record)
+        if self.should_colorize(record):
+            color = self.get_color(record)
+            if color:
+                rv = colorize(color, rv)
+        return rv
+
+
+class ColorizedStderrHandler(ColorizingStreamHandlerMixin, StderrHandler):
+    """A colorizing stream handler that writes to stderr.  It will only
+    colorize if a terminal was detected.  Note that this handler does
+    not colorize on Windows systems.
+
+    .. versionadded:: 0.3
+    """
+
+
+# backwards compat.  Should go away in some future releases
+from logbook.handlers import FingersCrossedHandler as \
+     FingersCrossedHandlerBase
+class FingersCrossedHandler(FingersCrossedHandlerBase):
+    def __init__(self, *args, **kwargs):
+        FingersCrossedHandlerBase.__init__(self, *args, **kwargs)
+        from warnings import warn
+        warn(PendingDeprecationWarning('fingers crossed handler changed '
+            'location.  It\'s now a core component of Logbook.'))
+
+
+class ExceptionHandler(Handler, StringFormatterHandlerMixin):
+    """An exception handler which raises exceptions of the given `exc_type`.
+    This is especially useful if you set a specific error `level` e.g. to treat
+    warnings as exceptions::
+
+        from logbook.more import ExceptionHandler
+
+        class ApplicationWarning(Exception):
+            pass
+
+        exc_handler = ExceptionHandler(ApplicationWarning, level='WARNING')
+
+    .. versionadded:: 0.3
+    """
+    def __init__(self, exc_type, level=NOTSET, format_string=None,
+                 filter=None, bubble=False):
+        Handler.__init__(self, level, filter, bubble)
+        StringFormatterHandlerMixin.__init__(self, format_string)
+        self.exc_type = exc_type
+
+    def handle(self, record):
+        if self.should_handle(record):
+            raise self.exc_type(self.format(record))
+        return False
